/* Put your custom CSS here */
@import url('http://fonts.cdnfonts.com/css/panton-black-caps');

h1 {
    font-family: "Panton Black Caps", sans-serif;
    color: #003B71 !important;
}

h2 {
    font-family: "Panton Light Caps", sans-serif;
    color: #322173 !important;
}

h3 {
    color: #322173 !important;
}

a {
    color: #5eb130 !important;
}



/* Fixing up some pygments and code-styling CSS for accessibility */
code { font-size: 100%; color: #e50051; }
pre { font-family: monospace; }

/* .highlight { font-size: 125%; } */
.highlight .c1 { color: #e50051; }
.highlight .si { color: #e50051; }
.highlight .nn { color: #e50051; }
